/**
 * Back 2 Browser Bytecode Translator
 * Copyright (C) 2012 Jaroslav Tulach <jaroslav.tulach@apidesign.org>
 *
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, version 2 of the License.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program. Look for COPYING file in the top folder.
 * If not, see http://opensource.org/licenses/GPL-2.0.
 */
package org.apidesign.bck2brwsr.htmlpage;

import java.lang.reflect.InvocationTargetException;
import java.lang.reflect.Method;
import java.util.ArrayList;
import java.util.Collection;
import java.util.Iterator;
import java.util.logging.Level;
import java.util.logging.Logger;
import org.apidesign.bck2brwsr.core.JavaScriptOnly;

/**
 *
 * @author Jaroslav Tulach <jtulach@netbeans.org>
 */
public final class KOList<T> extends ArrayList<T> {
    private final String name;
    private final String[] deps;
    private Knockout model;
    private Runnable onchange;

    public KOList(String name, String... deps) {
        this.name = name;
        this.deps = deps;
    }
    
    public void assign(Knockout model) {
        if (this.model != model) {
            this.model = model;
            notifyChange();
        }
    }
    
    public KOList<T> onChange(Runnable r) {
        if (this.onchange != null) {
            throw new IllegalStateException();
        }
        this.onchange = r;
        return this;
    }

    @Override
    public boolean add(T e) {
        boolean ret = super.add(e);
        notifyChange();
        return ret;
    }

    @Override
    public boolean addAll(Collection<? extends T> c) {
        boolean ret = super.addAll(c);
        notifyChange();
        return ret;
    }

    @Override
    public boolean addAll(int index, Collection<? extends T> c) {
        boolean ret = super.addAll(index, c);
        notifyChange();
        return ret;
    }

    @Override
    public boolean remove(Object o) {
        boolean ret = super.remove(o);
        notifyChange();
        return ret;
    }

    @Override
    public void clear() {
        super.clear();
        notifyChange();
    }

    @Override
    public boolean removeAll(Collection<?> c) {
        boolean ret = super.removeAll(c);
        notifyChange();
        return ret;
    }

    @Override
    public boolean retainAll(Collection<?> c) {
        boolean ret = super.retainAll(c);
        notifyChange();
        return ret;
    }

    @Override
    public T set(int index, T element) {
        T ret = super.set(index, element);
        notifyChange();
        return ret;
    }

    @Override
    public void add(int index, T element) {
        super.add(index, element);
        notifyChange();
    }

    @Override
    public T remove(int index) {
        T ret = super.remove(index);
        notifyChange();
        return ret;
    }

    @Override
    public String toString() {
        Iterator<T> it = iterator();
        if (!it.hasNext()) {
            return "[]";
        }
        String sep = "";
        StringBuilder sb = new StringBuilder();
        sb.append('[');
        while (it.hasNext()) {
            T t = it.next();
            sb.append(sep);
            sb.append(ConvertTypes.toJSON(t));
            sep = ",";
        }
        sb.append(']');
        return sb.toString();
    }
    
    
    @JavaScriptOnly(name = "koArray", value = "function() { return this.toArray___3Ljava_lang_Object_2(); }")
    private static native int koArray();
    
    public Object koData() {
        Object[] arr = toArray();
        Method toKO = null;
        for (int i = 0; i < arr.length; i++) {
            if (arr[i] == null) {
                continue;
            }
            if (toKO == null || toKO.getDeclaringClass() != arr[i].getClass()) {
                try {
                    toKO = arr[i].getClass().getDeclaredMethod("koData");
                    toKO.setAccessible(true);
                } catch (NoSuchMethodException ex) {
                    LOG.log(Level.FINE, "No koData conversion for " + arr[i], ex);
                    toKO = null;
                }
            }
            if (toKO != null) {
                try {
                    Object prev = arr[i];
                    arr[i] = toKO.invoke(arr[i]);
                    LOG.log(Level.FINER, "{0}th element {1} replaced by {2}", new Object[]{i, prev, arr[i]});
                } catch (IllegalAccessException | InvocationTargetException ex) {
                    LOG.log(Level.SEVERE, "Problems invoking koData on " + arr[i], ex);
                }
            }
        }
        return Knockout.toArray(arr);
    }
    private static final Logger LOG = Logger.getLogger(KOList.class.getName());

    private void notifyChange() {
        Knockout m = model;
        if (m != null) {
            m.valueHasMutated(name);
            for (String dependant : deps) {
                m.valueHasMutated(dependant);
            }
        }
        Runnable r = onchange;
        if (r != null) {
            r.run();
        }
    }

    @Override
    public KOList clone() {
        KOList ko = (KOList)super.clone();
        ko.model = null;
        return ko;
    }
    
}
